from pumapy.utilities.timer import Timer
from pumapy.utilities.workspace import Workspace
from pumapy.utilities.linear_solvers import PropertySolver
from scipy.sparse import csc_matrix
import numpy as np


class Permeability(PropertySolver):

    def __init__(self, workspace, solid_cutoff, tolerance, maxiter, solver_type, display_iter):
        allowed_solvers = ['minres', 'direct', 'cg', 'bicgstab']
        super().__init__(workspace, solver_type, allowed_solvers, tolerance, maxiter, display_iter)

        self.ws = workspace.copy()
        self.solid_cutoff = solid_cutoff
        self.ws.binarize_range(self.solid_cutoff)

        self.solver_type = solver_type
        self.len_x, self.len_y, self.len_z = self.ws.matrix.shape
        self.voxlength = self.ws.voxel_length

        self.nel = self.len_x * self.len_y
        self.nels = self.len_x * self.len_y * self.len_z
        self.nnP2 = (self.len_x + 1) * (self.len_y + 1)
        self.velF = np.where(self.ws.matrix.ravel(order='F') == 0)[0]  # only fluid elements
        self.nelF = self.velF.shape[0]

        self.ke = np.zeros((24, 24), dtype=float)
        self.ge = np.zeros((24, 8), dtype=float)
        self.pe = np.zeros((8, 8), dtype=float)
        self.fe = np.zeros((8, 1), dtype=float)
        self.x_full = np.zeros((4 * self.nels, 3), dtype=float)
        self.mConectP = np.zeros((self.nels, 8), dtype=np.uint64)
        self.mgdlF = np.zeros((32, self.nelF), dtype=np.uint64)
        self.resolveF = None

        self.keff = np.zeros((3, 3))
        self.solve_time = -1
        self.u_x = np.zeros((self.len_x, self.len_y, self.len_z, 3))
        self.u_y = np.zeros((self.len_x, self.len_y, self.len_z, 3))
        self.u_z = np.zeros((self.len_x, self.len_y, self.len_z, 3))

    def compute(self):
        t = Timer()
        self.initialize()
        self.assemble_bvector()
        self.assemble_Amatrix()
        print("Time to assemble matrices: ", t.elapsed()); t.reset()
        self.solve()
        print("Time to solve: ", t.elapsed())
        self.compute_effective_coefficient()
        self.solve_time = t.elapsed()

    def initialize(self):
        print("Initializing indexing matrices ... ", flush=True, end='')

        # Matrix with element connectivity with Periodic boundary conditions (PBC)
        mConectP1 = np.zeros((self.nel, 4, self.len_z + 1), dtype=np.uint64)
        aux = 1
        for slice in range(self.len_z):
            mIdNosP = np.reshape(np.arange(aux, aux + self.nel, dtype=np.uint64), (self.len_x, self.len_y), order='F')
            mIdNosP = np.append(mIdNosP, mIdNosP[0][np.newaxis], axis=0)  # Numbering bottom nodes
            mIdNosP = np.append(mIdNosP, mIdNosP[:, 0][:, np.newaxis], axis=1)  # Numbering right nodes
            mConectP1[:, 0, slice] = np.ravel(mIdNosP[1:, :-1], order='F')
            mConectP1[:, 1, slice] = np.ravel(mIdNosP[1:, 1:], order='F')
            mConectP1[:, 2, slice] = np.ravel(mIdNosP[:-1, 1:], order='F')
            mConectP1[:, 3, slice] = np.ravel(mIdNosP[:-1, :-1], order='F')
            aux += self.nel
        mConectP1[:, :, -1] = mConectP1[:, :, 0]

        for slice in range(self.len_z):
            self.mConectP[self.nel * slice:self.nel * (slice + 1), :4] = mConectP1[:, :, slice]
            self.mConectP[self.nel * slice:self.nel * (slice + 1), 4:] = mConectP1[:, :, slice + 1]

        self.calculate_element_matrices()

        # degrees of freedom matrix
        self.mgdlF[0] =  self.mConectP[self.velF, 0] * 3 - 2
        self.mgdlF[1] =  self.mConectP[self.velF, 0] * 3 - 1
        self.mgdlF[2] =  self.mConectP[self.velF, 0] * 3
        self.mgdlF[3] =  self.mConectP[self.velF, 1] * 3 - 2
        self.mgdlF[4] =  self.mConectP[self.velF, 1] * 3 - 1
        self.mgdlF[5] =  self.mConectP[self.velF, 1] * 3
        self.mgdlF[6] =  self.mConectP[self.velF, 2] * 3 - 2
        self.mgdlF[7] =  self.mConectP[self.velF, 2] * 3 - 1
        self.mgdlF[8] =  self.mConectP[self.velF, 2] * 3
        self.mgdlF[9] =  self.mConectP[self.velF, 3] * 3 - 2
        self.mgdlF[10] = self.mConectP[self.velF, 3] * 3 - 1
        self.mgdlF[11] = self.mConectP[self.velF, 3] * 3
        self.mgdlF[12] = self.mConectP[self.velF, 4] * 3 - 2
        self.mgdlF[13] = self.mConectP[self.velF, 4] * 3 - 1
        self.mgdlF[14] = self.mConectP[self.velF, 4] * 3
        self.mgdlF[15] = self.mConectP[self.velF, 5] * 3 - 2
        self.mgdlF[16] = self.mConectP[self.velF, 5] * 3 - 1
        self.mgdlF[17] = self.mConectP[self.velF, 5] * 3
        self.mgdlF[18] = self.mConectP[self.velF, 6] * 3 - 2
        self.mgdlF[19] = self.mConectP[self.velF, 6] * 3 - 1
        self.mgdlF[20] = self.mConectP[self.velF, 6] * 3
        self.mgdlF[21] = self.mConectP[self.velF, 7] * 3 - 2
        self.mgdlF[22] = self.mConectP[self.velF, 7] * 3 - 1
        self.mgdlF[23] = self.mConectP[self.velF, 7] * 3
        self.mgdlF[24] = 3 * self.nels + self.mConectP[self.velF, 0]
        self.mgdlF[25] = 3 * self.nels + self.mConectP[self.velF, 1]
        self.mgdlF[26] = 3 * self.nels + self.mConectP[self.velF, 2]
        self.mgdlF[27] = 3 * self.nels + self.mConectP[self.velF, 3]
        self.mgdlF[28] = 3 * self.nels + self.mConectP[self.velF, 4]
        self.mgdlF[29] = 3 * self.nels + self.mConectP[self.velF, 5]
        self.mgdlF[30] = 3 * self.nels + self.mConectP[self.velF, 6]
        self.mgdlF[31] = 3 * self.nels + self.mConectP[self.velF, 7]
        print("Done")

    def assemble_bvector(self):
        print("Assembling b vector ... ", flush=True, end='')
        iF = np.hstack((np.reshape(self.mgdlF[:24:3],  self.nelF * 8, order='F'),
                        np.reshape(self.mgdlF[1:24:3], self.nelF * 8, order='F'),
                        np.reshape(self.mgdlF[2:24:3], self.nelF * 8, order='F'))) - 1
        jF = np.hstack((np.zeros(self.nelF * 8, dtype=np.uint8),
                        np.ones(self.nelF * 8, dtype=np.uint8),
                        np.full(self.nelF * 8, 2, dtype=np.uint8)))
        sF = np.squeeze(np.tile(self.fe, (self.nelF * 3, 1)))
        self.bvec_full = csc_matrix((sF, (iF, jF)), shape=(4 * self.nels, 3))
        print("Done")

    def assemble_Amatrix(self):
        print("Initializing large data structures ... ", flush=True, end='')
        iK = np.repeat(np.reshape(self.mgdlF[:24], self.nelF * 24, order='F'), 24)
        jK = np.reshape(np.repeat(self.mgdlF[:24], 24, axis=1), self.nelF * 576, order='F')
        iG = np.repeat(np.reshape(self.mgdlF[:24], self.nelF * 24, order='F'), 8)
        jG = np.reshape(np.repeat(self.mgdlF[24:], 24, axis=1), self.nelF * 192, order='F')
        iP = np.repeat(np.reshape(self.mgdlF[24:], self.nelF * 8, order='F'), 8)
        jP = np.reshape(np.repeat(self.mgdlF[24:], 8, axis=1), self.nelF * 64, order='F')
        del self.mgdlF
        iA = np.hstack((iK, iG, jG, iP)) - 1
        jA = np.hstack((jK, jG, iG, jP)) - 1
        coeff = np.hstack((np.tile(self.ke, self.nelF), np.tile(self.ge, self.nelF),
                           np.tile(self.ge, self.nelF), -np.tile(self.pe, self.nelF)))

        print("Done\nAssembling A matrix ... ", flush=True, end='')
        self.Amat = csc_matrix((coeff, (iA, jA)))
        print("Done")

        print("Reducing system of equations ... ", flush=True, end='')
        resolveF_u = np.arange(1, self.nels * 3 + 1, dtype=np.uint64)
        nosnulos = np.unique(self.mConectP[np.where(self.ws.matrix.ravel(order='F') > 0)[0], :8])
        gdlnulos = np.hstack((nosnulos * 3 - 2, nosnulos * 3 - 1, nosnulos * 3))
        resolveF_u = np.delete(resolveF_u, gdlnulos - 1)
        resolveF_p = self.nels * 3 + np.unique(self.mConectP[np.where(self.ws.matrix.ravel(order='F') == 0)[0], :8])
        self.resolveF = np.hstack((resolveF_u, resolveF_p)) - 1
        del self.mConectP

        self.Amat = self.Amat[self.resolveF][:, self.resolveF]
        self.bvec_full = self.bvec_full[self.resolveF]
        print("Done")

    def solve(self):
        # solves the system in the 3 directions directly
        if self.solver_type == 'direct':
            self.bvec = self.bvec_full
            super().solve()
            self.x_full[self.resolveF] = self.x.toarray()
        else:  # solves one direction at a time
            self.del_matrices = False
            for i in range(3):
                self.bvec = self.bvec_full[:, i]
                super().solve()
                self.x_full[self.resolveF, i] = self.x
            del self.Amat, self.bvec, self.initial_guess

    def compute_effective_coefficient(self):
        aux = 1
        vIdNosP = np.zeros(self.len_z * self.nnP2, dtype=np.uint64)
        for slice in range(self.len_z):
            mIdNosP = np.reshape(np.arange(aux, aux + self.nel, dtype=np.uint64), (self.len_x, self.len_y), order='F')
            mIdNosP = np.append(mIdNosP, mIdNosP[0][np.newaxis], axis=0)  # Numbering bottom nodes
            mIdNosP = np.append(mIdNosP, mIdNosP[:, 0][:, np.newaxis], axis=1)  # Numbering right nodes

            vIdNosP[self.nnP2 * slice:self.nnP2 * (slice + 1)] = mIdNosP.ravel(order='F')
            aux += self.nel
        del mIdNosP

        vIdNosP = np.append(vIdNosP, (vIdNosP[:self.nnP2]))
        vIdNosP = np.unique(vIdNosP)

        self.keff[0, 0] =   np.sum(self.x_full[vIdNosP * 3 - 2, 1])
        self.keff[0, 1] = - np.sum(self.x_full[vIdNosP * 3 - 3, 1])
        self.keff[0, 2] = - np.sum(self.x_full[vIdNosP * 3 - 1, 1])
        self.keff[1, 0] = - np.sum(self.x_full[vIdNosP * 3 - 2, 0])
        self.keff[1, 1] =   np.sum(self.x_full[vIdNosP * 3 - 3, 0])
        self.keff[1, 2] =   np.sum(self.x_full[vIdNosP * 3 - 1, 0])
        self.keff[2, 0] = - np.sum(self.x_full[vIdNosP * 3 - 2, 2])
        self.keff[2, 1] =   np.sum(self.x_full[vIdNosP * 3 - 3, 2])
        self.keff[2, 2] =   np.sum(self.x_full[vIdNosP * 3 - 1, 2])
        self.keff /= self.nels
        print(f'\nEffective permeability tensor: \n{self.keff}')

        # Extracting velocity fields
        self.u_x[:, :, :, 0] =   np.reshape(self.x_full[vIdNosP * 3 - 2, 1], (self.len_x, self.len_y, self.len_z), order='F')
        self.u_x[:, :, :, 1] = - np.reshape(self.x_full[vIdNosP * 3 - 3, 1], (self.len_x, self.len_y, self.len_z), order='F')
        self.u_x[:, :, :, 2] = - np.reshape(self.x_full[vIdNosP * 3 - 1, 1], (self.len_x, self.len_y, self.len_z), order='F')
        self.u_y[:, :, :, 0] = - np.reshape(self.x_full[vIdNosP * 3 - 2, 0], (self.len_x, self.len_y, self.len_z), order='F')
        self.u_y[:, :, :, 1] =   np.reshape(self.x_full[vIdNosP * 3 - 3, 0], (self.len_x, self.len_y, self.len_z), order='F')
        self.u_y[:, :, :, 2] =   np.reshape(self.x_full[vIdNosP * 3 - 1, 0], (self.len_x, self.len_y, self.len_z), order='F')
        self.u_z[:, :, :, 0] = - np.reshape(self.x_full[vIdNosP * 3 - 2, 2], (self.len_x, self.len_y, self.len_z), order='F')
        self.u_z[:, :, :, 1] =   np.reshape(self.x_full[vIdNosP * 3 - 3, 2], (self.len_x, self.len_y, self.len_z), order='F')
        self.u_z[:, :, :, 2] =   np.reshape(self.x_full[vIdNosP * 3 - 1, 2], (self.len_x, self.len_y, self.len_z), order='F')

    def calculate_element_matrices(self):
        coordsElem = np.array([[0, 0, 0], [self.voxlength, 0, 0], [self.voxlength, self.voxlength, 0],
                               [0, self.voxlength, 0], [0, 0, self.voxlength], [self.voxlength, 0, self.voxlength],
                               [self.voxlength, self.voxlength, self.voxlength], [0, self.voxlength, self.voxlength]])
        rr = np.array([-1. / np.sqrt(3), 1. / np.sqrt(3)])
        ss = rr.copy()
        tt = rr.copy()
        ww = np.array([1, 1])
        C = np.diag([2., 2., 2., 1., 1., 1.])
        stab = (self.voxlength ** 2 + self.voxlength ** 2 + self.voxlength ** 2) / 18.
        mat111000 = np.array([[1.], [1.], [1.], [0.], [0.], [0.]])

        for i in range(2):
            r = rr[i]
            for j in range(2):
                s = ss[j]
                for k in range(2):
                    t = tt[k]

                    N1 = (1 - r) * (1 - s) * (1 - t)
                    N2 = (1 + r) * (1 - s) * (1 - t)
                    N3 = (1 + r) * (1 + s) * (1 - t)
                    N4 = (1 - r) * (1 + s) * (1 - t)
                    N5 = (1 - r) * (1 - s) * (1 + t)
                    N6 = (1 + r) * (1 - s) * (1 + t)
                    N7 = (1 + r) * (1 + s) * (1 + t)
                    N8 = (1 - r) * (1 + s) * (1 + t)
                    N = 0.125 * np.array([[N1, 0, 0, N2, 0, 0, N3, 0, 0, N4, 0, 0, N5, 0, 0, N6, 0, 0, N7, 0, 0, N8, 0, 0],
                                          [0, N1, 0, 0, N2, 0, 0, N3, 0, 0, N4, 0, 0, N5, 0, 0, N6, 0, 0, N7, 0, 0, N8, 0],
                                          [0, 0, N1, 0, 0, N2, 0, 0, N3, 0, 0, N4, 0, 0, N5, 0, 0, N6, 0, 0, N7, 0, 0, N8]])

                    dN1dr = -0.125 * (1 - s) * (1 - t)
                    dN1ds = -0.125 * (1 - r) * (1 - t)
                    dN1dt = -0.125 * (1 - r) * (1 - s)
                    dN2dr = +0.125 * (1 - s) * (1 - t)
                    dN2ds = -0.125 * (1 + r) * (1 - t)
                    dN2dt = -0.125 * (1 + r) * (1 - s)
                    dN3dr = +0.125 * (1 + s) * (1 - t)
                    dN3ds = +0.125 * (1 + r) * (1 - t)
                    dN3dt = -0.125 * (1 + r) * (1 + s)
                    dN4dr = -0.125 * (1 + s) * (1 - t)
                    dN4ds = +0.125 * (1 - r) * (1 - t)
                    dN4dt = -0.125 * (1 - r) * (1 + s)
                    dN5dr = -0.125 * (1 - s) * (1 + t)
                    dN5ds = -0.125 * (1 - r) * (1 + t)
                    dN5dt = +0.125 * (1 - r) * (1 - s)
                    dN6dr = +0.125 * (1 - s) * (1 + t)
                    dN6ds = -0.125 * (1 + r) * (1 + t)
                    dN6dt = +0.125 * (1 + r) * (1 - s)
                    dN7dr = +0.125 * (1 + s) * (1 + t)
                    dN7ds = +0.125 * (1 + r) * (1 + t)
                    dN7dt = +0.125 * (1 + r) * (1 + s)
                    dN8dr = -0.125 * (1 + s) * (1 + t)
                    dN8ds = +0.125 * (1 - r) * (1 + t)
                    dN8dt = +0.125 * (1 - r) * (1 + s)
                    DN = np.array([[dN1dr, dN2dr, dN3dr, dN4dr, dN5dr, dN6dr, dN7dr, dN8dr],
                                   [dN1ds, dN2ds, dN3ds, dN4ds, dN5ds, dN6ds, dN7ds, dN8ds],
                                   [dN1dt, dN2dt, dN3dt, dN4dt, dN5dt, dN6dt, dN7dt, dN8dt]])

                    J = DN @ coordsElem
                    detJ = np.linalg.det(J)
                    weight = detJ * ww[i] * ww[j] * ww[k]
                    invJ = np.linalg.inv(J)
                    DNxy = invJ @ DN
                    B = np.array([[DNxy[0, 0], 0, 0, DNxy[0, 1], 0, 0, DNxy[0, 2], 0, 0, DNxy[0, 3], 0, 0, DNxy[0, 4], 0, 0, DNxy[0, 5], 0, 0, DNxy[0, 6], 0, 0, DNxy[0, 7], 0, 0],
                                  [0, DNxy[1, 0], 0, 0, DNxy[1, 1], 0, 0, DNxy[1, 2], 0, 0, DNxy[1, 3], 0, 0, DNxy[1, 4], 0, 0, DNxy[1, 5], 0, 0, DNxy[1, 6], 0, 0, DNxy[1, 7], 0],
                                  [0, 0, DNxy[2, 0], 0, 0, DNxy[2, 1], 0, 0, DNxy[2, 2], 0, 0, DNxy[2, 3], 0, 0, DNxy[2, 4], 0, 0, DNxy[2, 5], 0, 0, DNxy[2, 6], 0, 0, DNxy[2, 7]],
                                  [DNxy[1, 0], DNxy[0, 0], 0, DNxy[1, 1], DNxy[0, 1], 0, DNxy[1, 2], DNxy[0, 2], 0, DNxy[1, 3], DNxy[0, 3], 0,
                                   DNxy[1, 4], DNxy[0, 4], 0, DNxy[1, 5], DNxy[0, 5], 0, DNxy[1, 6], DNxy[0, 6], 0, DNxy[1, 7], DNxy[0, 7], 0],
                                  [DNxy[2, 0], 0, DNxy[0, 0], DNxy[2, 1], 0, DNxy[0, 1], DNxy[2, 2], 0, DNxy[0, 2], DNxy[2, 3], 0, DNxy[0, 3],
                                   DNxy[2, 4], 0, DNxy[0, 4], DNxy[2, 5], 0, DNxy[0, 5], DNxy[2, 6], 0, DNxy[0, 6], DNxy[2, 7], 0, DNxy[0, 7]],
                                  [0, DNxy[2, 0], DNxy[1, 0], 0, DNxy[2, 1], DNxy[1, 1], 0, DNxy[2, 2], DNxy[1, 2], 0, DNxy[2, 3], DNxy[1, 3],
                                   0, DNxy[2, 4], DNxy[1, 4], 0, DNxy[2, 5], DNxy[1, 5], 0, DNxy[2, 6], DNxy[1, 6], 0, DNxy[2, 7], DNxy[1, 7]]])

                    self.ke += weight * B.T @ C @ B
                    self.ge += weight * B.T @ mat111000 @ N.ravel(order='F')[::9][np.newaxis]
                    self.pe += weight * stab * DNxy.T @ DNxy
                    self.fe += weight * N[0][::3][:, np.newaxis]

        self.ke = self.ke.ravel()
        self.ge = self.ge.ravel()
        self.pe = self.pe.ravel()

    def log_input(self):
        self.ws.log.log_section("Computing Permeability")
        self.ws.log.log_line("Domain Size: " + str(self.ws.get_shape()))
        self.ws.log.log_line("Solver Type: " + str(self.solver_type))
        self.ws.log.log_line("Solver Tolerance: " + str(self.tolerance))
        self.ws.log.log_line("Max Iterations: " + str(self.maxiter))
        self.ws.log.write_log()

    def log_output(self):
        self.ws.log.log_section("Finished Permeability Calculation")
        self.ws.log.log_line("Permeability: " + str(self.keff))
        self.ws.log.log_line("Solver Time: " + str(self.solve_time))
        self.ws.log.write_log()

    def error_check(self):
        if not isinstance(self.ws, Workspace):
            raise Exception("Workspace must be a puma.Workspace.")



import unittest
import numpy as np
import pumapy as puma


class TestAnisotropicTC(unittest.TestCase):

    def setUp(self):
        self.X = 10
        self.Y = 8
        self.Z = 14

        self.ws_homoiso = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        self.cond_map_homoiso = puma.AnisotropicConductivityMap()
        self.cond_map_homoiso.add_material((1, 1), 10, 11, 12, 0.5, 0.2, 0.4)
        # conductivity solution tensor for homogeneous
        # 10  0.5 0.2
        # 0.5  11 0.4
        # 0.2 0.4, 12

        self.solution_homoiso_x = np.zeros((self.X, self.Y, self.Z), dtype=float)
        for i in range(self.X):
            self.solution_homoiso_x[i] = i / (self.X - 1.)

        self.solution_homoiso_y = np.zeros((self.X, self.Y, self.Z), dtype=float)
        for j in range(self.Y):
            self.solution_homoiso_y[:, j] = j / (self.Y - 1.)

        self.solution_homoiso_z = np.zeros((self.X, self.Y, self.Z), dtype=float)
        for k in range(self.Z):
            self.solution_homoiso_z[:, :, k] = k / (self.Z - 1.)

        self.ws_matSeriesInx = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        self.ws_matSeriesInx[int(self.ws_matSeriesInx.matrix.shape[0] / 2.):, :] = 2  # in series in x

        self.ws_matSeriesIny = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        self.ws_matSeriesIny[:, int(self.ws_matSeriesIny.matrix.shape[1] / 2.):] = 2  # in series in y

        self.ws_matSeriesInz = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        self.ws_matSeriesInz[:, :, int(self.ws_matSeriesInz.matrix.shape[2] / 2.):] = 2  # in series in z

        self.cond_map_matSeries = puma.AnisotropicConductivityMap()
        self.cond_map_matSeries.add_material((1, 1), 10, 10, 10, 0.2, 0.3, 0.5)
        self.cond_map_matSeries.add_material((2, 2), 1, 1, 1, 0.2, 0.3, 0.5)
        # conductivity solution tensor for mat in series along x
        # 1.8181 0.2 0.3
        #    0.2 5.5 0.5
        #    0.3 0.5 5.5

        # conductivity solution tensor for mat in series along y
        # 5.5    0.2 0.3
        # 0.2 1.8181 0.5
        # 0.3    0.5 5.5

        # conductivity solution tensor for mat in series along z
        # 5.5 0.2    0.3
        # 0.2 5.5    0.5
        # 0.3 0.5 1.8181

    def test_homoiso_x(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_homoiso, self.cond_map_homoiso, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [10, 0.5, 0.2])
        np.testing.assert_array_almost_equal(T, self.solution_homoiso_x)

    def test_homoiso_y(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_homoiso, self.cond_map_homoiso, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.5, 11, 0.4])
        np.testing.assert_array_almost_equal(T, self.solution_homoiso_y)

    def test_homoiso_z(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_homoiso, self.cond_map_homoiso, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.2, 0.4, 12])
        np.testing.assert_array_almost_equal(T, self.solution_homoiso_z)

    def test_matSeriesInx_x_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [1.818181818, 0.2, 0.3])

    def test_matSeriesInx_y_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.2, 5.5, 0.5])

    def test_matSeriesInx_z_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.3, 0.5, 5.5])

    def test_matSeriesIny_x_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesIny, self.cond_map_matSeries, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [5.5, 0.2, 0.3])

    def test_matSeriesIny_y_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesIny, self.cond_map_matSeries, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.2, 1.818181818, 0.5])

    def test_matSeriesIny_z_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesIny, self.cond_map_matSeries, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.3, 0.5, 5.5])

    def test_matSeriesInz_x_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInz, self.cond_map_matSeries, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [5.5, 0.2, 0.3])

    def test_matSeriesInz_y_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInz, self.cond_map_matSeries, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.2, 5.5, 0.5])

    def test_matSeriesInz_z_per(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInz, self.cond_map_matSeries, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.3, 0.5, 1.818181818])

    def test_matSeriesInx_x_sym(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'x', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [1.818181818, 0.2, 0.3])

    def test_matSeriesInx_y_sym(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'y', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.2, 5.5, 0.5])

    def test_matSeriesInx_z_sym(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'z', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0.3, 0.5, 5.5])

    def test_matSeriesInx_x_sym_bicgstab(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'x', 's', solver_type='bicgstab')
        np.testing.assert_array_almost_equal(keff, [1.818181818, 0.2, 0.3], decimal=4)

    def test_matSeriesInx_y_sym_bicgstab(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'y', 's', solver_type='bicgstab')
        np.testing.assert_array_almost_equal(keff, [0.2, 5.5, 0.5], decimal=4)

    def test_matSeriesInx_z_sym_bicgstab(self):
        keff, T, _ = puma.compute_thermal_conductivity(self.ws_matSeriesInx, self.cond_map_matSeries, 'z', 's', solver_type='bicgstab')
        np.testing.assert_array_almost_equal(keff, [0.3, 0.5, 5.5], decimal=4)

    def test_tensor_rotation_x(self):
        ws = puma.Workspace.from_array(np.zeros((self.X, self.Y, self.Z)))
        ws.set(1, (1, 0, 0))
        cond_map = puma.AnisotropicConductivityMap()
        cond_map.add_material_to_orient((1, 1), 10, 1)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'x', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [10, 0, 0], decimal=6)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'y', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0, 1, 0], decimal=6)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'z', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0, 0, 1], decimal=6)

    def test_tensor_rotation_y(self):
        ws = puma.Workspace.from_array(np.zeros((self.X, self.Y, self.Z)))
        ws.set(1, (0, 1, 0))
        cond_map = puma.AnisotropicConductivityMap()
        cond_map.add_material_to_orient((1, 1), 10, 1)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'x', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [1, 0, 0], decimal=6)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'y', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0, 10, 0], decimal=6)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'z', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0, 0, 1], decimal=6)

    def test_tensor_rotation_z(self):
        ws = puma.Workspace.from_array(np.zeros((self.X, self.Y, self.Z)))
        ws.set(1, (0, 0, 1))
        cond_map = puma.AnisotropicConductivityMap()
        cond_map.add_material_to_orient((1, 1), 10, 1)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'x', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [1, 0, 0], decimal=6)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'y', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0, 1, 0], decimal=6)

        keff, T, q = puma.compute_thermal_conductivity(ws, cond_map, 'z', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(keff, [0, 0, 10], decimal=6)

    def test_artfib50(self):
        ws = puma.import_vti("testdata/artifib.vtk")
        cond_map = puma.AnisotropicConductivityMap()
        cond_map.add_isotropic_material((0, 128), 0.0257)
        cond_map.add_isotropic_material((129, 255), 10)
        keff, _, _ = puma.compute_thermal_conductivity(ws, cond_map, 'x', 's', tolerance=1e-4)
        np.testing.assert_array_almost_equal(keff, [0.0982, -0.0029, 0.0034], decimal=4)
        keff, _, _ = puma.compute_thermal_conductivity(ws, cond_map, 'y', 's', tolerance=1e-4)
        np.testing.assert_array_almost_equal(keff, [-0.00188, 0.16165, 0.001233], decimal=4)
        keff, _, _ = puma.compute_thermal_conductivity(ws, cond_map, 'z', 's', tolerance=1e-4)
        np.testing.assert_array_almost_equal(keff, [0.001, -0.0034, 0.0848], decimal=4)


if __name__ == '__main__':
    unittest.main()



import datetime
import os


def print_warning(warning_text):
    print(Colors.WARNING + warning_text + Colors.ENDC)


class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Logger:

    def __init__(self):

        self.log = ""
        self.location = ""

        if not os.path.exists('logs'):
            os.mkdir('logs')

        now = datetime.datetime.now()
        month = str(now.month) if len(str(now.month)) == 2 else "0" + str(now.month)
        day = str(now.day) if len(str(now.day)) == 2 else "0" + str(now.day)
        hour = str(now.hour) if len(str(now.hour)) == 2 else "0" + str(now.hour)
        minute = str(now.minute) if len(str(now.minute)) == 2 else "0" + str(now.minute)
        second = str(now.second) if len(str(now.second)) == 2 else "0" + str(now.second)
        self.location = "logs/pumapy_log_" + str(now.year) + month + day + "_" + hour + minute + second + ".txt"

    def set_location(self, location):
        self.location = location

    def log_section(self, name):
        self.log += "\n---------------------------- \n " + str(name) + "\n---------------------------- \n"

    def log_line(self, val):
        self.log += str(val) + "\n"

    def log_item(self, val):
        self.log += str(val) + " "

    def new_line(self):
        self.log += "\n"

    def log_bool(self, var_name, val):
        if val:
            self.log_line(str(var_name) + ": True")
        else:
            self.log_line(str(var_name) + ": False")

    def log_value(self, var_name, val):
        self.log_line(str(var_name) + ": " + str(val))

    def write_log(self):
        if self.location == "":
            return
        txt_file = open(self.location, "w")
        txt_file.write(self.log)
        txt_file.close()



from skimage.io import imread
import numpy as np
import pickle
from pumapy import Workspace
from os import path
from glob import glob
from vtkmodules.util.numpy_support import vtk_to_numpy
from vtkmodules.vtkIOXML import vtkXMLImageDataReader, vtkXMLUnstructuredGridReader
from vtkmodules.vtkIOLegacy import vtkDataSetReader


def import_3Dtiff(filename, voxel_length=1e-6, import_ws=True):
    """ Function to io 3D tiff

    :param filename: filepath and name
    :type filename: string
    :param voxel_length: size of a voxel side
    :type voxel_length: float, optional
    :param import_ws: if True returns a puma.Workspace, otherwise a ndarray
    :type import_ws: bool, optional
    :return: domain
    :rtype: Workspace or ndarray

    :Example:
    >>> import pumapy as puma
    >>> ws_tiff = puma.import_3Dtiff(puma.path_to_example_file("50_artfibers.tif"), 1.3e-6, import_ws=True)
    >>> ws_tiff.get_shape()
    """
    print("Importing " + filename + " ... ", end='')

    if not path.exists(filename):
        raise Exception("File " + filename + " not found.")

    nparray = imread(filename).astype(np.uint16)

    if nparray.ndim == 2:
        nparray = nparray[:, :, np.newaxis]
    else:
        nparray = nparray.transpose(2, 1, 0)
    print("Done")

    if import_ws:
        ws = Workspace.from_array(nparray)
        ws.set_voxel_length(voxel_length)
        return ws
    return nparray


def import_bin(filename):
    """ Import a puma.Workspace to binary (.pumapy extension)

    :param filename: filepath and name
    :type filename: string
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws_binary = puma.import_bin(puma.path_to_example_file("fibers_with_orientation.pumapy"))
    """
    print("Importing " + filename + " ... ", end='')

    if not path.exists(filename):
        raise Exception("File " + filename + " not found.")

    pumapy_file = open(filename, 'rb')
    ws = pickle.load(pumapy_file)
    pumapy_file.close()
    print("Done")
    return ws


def import_vti(filename, voxel_length=None, import_ws=True):
    """ Function to import either legacy VTK file (.vtk) or vtkImageData (.vti)

    :param filename: filepath and name
    :type filename: string
    :param voxel_length: voxel_length. If None, voxel_length from the vtk file is used
    :type voxel_length: float, optional
    :param import_ws: True returns a puma.Workspace, otherwise a list of ndarrays
    :type import_ws: bool, optional
    :return: if import_ws is True, then it returns a Workspace.
        if import_ws is False, it returns a dictionary of ndarrays as {"name1": data1, "name2": data2 ...}
    :rtype: Workspace or dict(string: ndarray)

    :Example:
    >>> import pumapy as puma
    >>> ws_vtk = puma.import_vti(puma.path_to_example_file("fibers_with_orientation.vti"))
    """
    print("Importing " + filename + " ... ", end='')

    if not path.exists(filename):
        raise Exception("File " + filename + " not found.")

    if filename[-4:] == ".vti":
        reader = vtkXMLImageDataReader()
        reader.SetFileName(filename)
        reader.Update()
        vtkobject = reader.GetOutput()

    elif filename[-4:] == ".vtk":
        reader = vtkDataSetReader()
        reader.SetFileName(filename)
        reader.ReadAllScalarsOn()
        reader.ReadAllColorScalarsOn()
        reader.ReadAllNormalsOn()
        reader.ReadAllTCoordsOn()
        reader.ReadAllVectorsOn()
        reader.Update()  # reading
        vtkobject = reader.GetOutputDataObject(0)

    else:
        raise Exception("File of an unrecognized extension, only .vti and .vtk supported.")

    shape = vtkobject.GetDimensions()
    orientation = None

    if vtkobject.GetPointData().GetNumberOfArrays() == 0 and vtkobject.GetCellData().GetNumberOfArrays() == 0:
        raise Exception("No CELL_DATA or POINT_DATA arrays detected in file.")

    if import_ws:

        if vtkobject.GetCellData().GetNumberOfArrays() + vtkobject.GetPointData().GetNumberOfArrays() > 2:
            raise Exception("More than two arrays in file detected: set the import_ws to False to import it.")

        # checking for CELL_DATA
        if vtkobject.GetCellData().GetNumberOfArrays() > 0:
            if vtkobject.GetCellData().GetNumberOfArrays() == 1:  # if one array, could either be orientation or matrix
                nparray = vtk_to_numpy(vtkobject.GetCellData().GetArray(0))
                if nparray.ndim == 2:  # if orientation
                    orientation = nparray.copy()
                    nparray = None
            else:
                nparray = vtk_to_numpy(vtkobject.GetCellData().GetArray(0))
                orientation = vtk_to_numpy(vtkobject.GetCellData().GetArray(1))

        # checking for POINT_DATA
        else:
            if vtkobject.GetPointData().GetNumberOfArrays() == 1:  # if one array, could either be orientation or matrix
                nparray = vtk_to_numpy(vtkobject.GetPointData().GetArray(0))
                if nparray.ndim == 2:  # if orientation
                    orientation = nparray.copy()
                    nparray = None
            else:
                nparray = vtk_to_numpy(vtkobject.GetPointData().GetArray(0))
                orientation = vtk_to_numpy(vtkobject.GetPointData().GetArray(1))

        if nparray is not None:
            nparray = nparray.reshape(shape[0] - 1, shape[1] - 1, shape[2] - 1, order="F")
        if orientation is not None:
            orientation = orientation.reshape(shape[0] - 1, shape[1] - 1, shape[2] - 1, 3, order="F")
        print("Done")

        if nparray is None:
            ws = Workspace.from_shape((shape[0] - 1, shape[1] - 1, shape[2] - 1))
        else:
            ws = Workspace.from_array(nparray)
        if orientation is not None:
            ws.set_orientation(orientation)
        if voxel_length is None:
            ws.set_voxel_length(vtkobject.GetSpacing()[0])
        else:
            ws.set_voxel_length(voxel_length)
        return ws

    else:
        nparray_list = dict()
        for i in range(vtkobject.GetCellData().GetNumberOfArrays()):
            tmp = vtk_to_numpy(vtkobject.GetCellData().GetArray(i))
            if tmp.ndim == 1:
                tmp = tmp.reshape(shape[0] - 1, shape[1] - 1, shape[2] - 1, order="F")
            else:
                tmp = tmp.reshape(shape[0] - 1, shape[1] - 1, shape[2] - 1, 3, order="F")
            nparray_list[vtkobject.GetCellData().GetArrayName(i)] = tmp
        for i in range(vtkobject.GetPointData().GetNumberOfArrays()):
            tmp = vtk_to_numpy(vtkobject.GetPointData().GetArray(i))
            if tmp.ndim == 1:
                tmp = tmp.reshape(shape[0] - 1, shape[1] - 1, shape[2] - 1, order="F")
            else:
                tmp = tmp.reshape(shape[0] - 1, shape[1] - 1, shape[2] - 1, 3, order="F")
            nparray_list[vtkobject.GetPointData().GetArrayName(i)] = tmp
        print("Done")
        return nparray_list


def import_weave_vtu(filename):
    """ Import TexGen vtu weave in a Workspace

    :param filename: file path and name
    :type filename: string
    :return: voxelized weave from TexGen
    :rtype: Workspace
    """

    if not path.exists(filename):
        filename = glob(filename + '*.vtu')
        if len(filename) == 0:
            raise Exception("File " + filename + " not found.")
        else:
            filename = filename[0]
    print("Importing " + filename + " ... ", end='')

    reader = vtkXMLUnstructuredGridReader()
    reader.SetFileName(filename)
    reader.Update()  # reading
    vtkobject = reader.GetOutputDataObject(0)

    dims = path.split(filename[:-4])[1].split('_')[-3:]

    # ORIGINAL TEXGEN
    # Number Of Arrays: 6
    # Array 0 name = YarnIndex  <-- transfering to ws
    # Array 1 name = YarnTangent
    # Array 2 name = Location
    # Array 3 name = VolumeFraction
    # Array 4 name = SurfaceDistance
    # Array 5 name = Orientation  <-- transfering to ws
    # yarn_index = vtk_to_numpy(vtkobject.GetCellData().GetArray(0)) + 1
    # orientation = vtk_to_numpy(vtkobject.GetCellData().GetArray(5))
    # yarn_index = yarn_index.reshape(int(dims[0]), int(dims[1]), int(dims[2]), order="F")
    # orientation = orientation.reshape(int(dims[0]), int(dims[1]), int(dims[2]), 3, order="F")

    # MODIFIED TEXGEN
    # Number Of Arrays: 6
    # Array 0 name = YarnIndex  <-- transfering to ws
    # Array 1 name = Orientation  <-- depends on has_orientation
    yarn_index = vtk_to_numpy(vtkobject.GetCellData().GetArray(0)) + 1
    yarn_index = yarn_index.reshape(int(dims[0]), int(dims[1]), int(dims[2]), order="F")
    ws = Workspace.from_array(yarn_index)

    if vtkobject.GetCellData().GetNumberOfArrays() == 2:
        orientation = vtk_to_numpy(vtkobject.GetCellData().GetArray(1))
        orientation = orientation.reshape(int(dims[0]), int(dims[1]), int(dims[2]), 3, order="F")
        ws.set_orientation(orientation)

    print("Done")
    return ws



from pumapy import Workspace
import numpy as np


class ConductivityBC:
    def __init__(self, x, y, z):
        self.dirichlet = np.full((x, y, z), np.Inf, dtype=float)
        # self.__neumann = np.full((x, y, z, 3), np.Inf, dtype=float)

    def __getitem__(self, key):
        return self.dirichlet[key]

    def __setitem__(self, key, value):
        self.dirichlet[key] = value

    @classmethod
    def from_workspace(cls, workspace):
        """ Generate ConductivityBC from a Workspace.

        :param workspace: domain
        :type workspace: Workspace
        """
        if isinstance(workspace, Workspace):
            return cls(workspace.len_x(), workspace.len_y(), workspace.len_z())
        else:
            raise Exception("Cannot create ConductivityBC, the input to the from_workspace method has to be a Workspace.")

    def get_shape(self):
        return self.dirichlet.shape

    def show(self):
        Workspace.show_matrix(self.dirichlet)


class Isotropic_symmetricBC:
    def __init__(self, len_x, len_y, len_z):
        self.len_x = len_x
        self.len_y = len_y
        self.len_z = len_z
        self.len_xy = self.len_x * self.len_y

    def indexAt_3D(self, i, j, k):
        if i == -1:
            i = 0
        if i == self.len_x:
            i = self.len_x - 1
        if j == -1:
            j = 0
        if j == self.len_y:
            j = self.len_y - 1
        if k == -1:
            k = 0
        if k == self.len_z:
            k = self.len_z - 1
        return self.len_xy * k + self.len_x * j + i


class Isotropic_periodicBC:
    def __init__(self, len_x, len_y, len_z):
        self.len_x = len_x
        self.len_y = len_y
        self.len_z = len_z
        self.len_xy = self.len_x * self.len_y

    def indexAt_3D(self, i, j, k):
        if i == -1:
            i = self.len_x - 1
        if i == self.len_x:
            i = 0
        if j == -1:
            j = self.len_y - 1
        if j == self.len_y:
            j = 0
        if k == -1:
            k = self.len_z - 1
        if k == self.len_z:
            k = 0
        return self.len_xy * k + self.len_x * j + i


class ElasticityBC:
    def __init__(self, x, y, z):
        self.dirichlet = np.full((x, y, z, 3), np.Inf, dtype=float)
        # self.__neumann = np.full((x, y, z, 6), np.Inf, dtype=float)

    def __getitem__(self, key):
        return self.dirichlet[key]

    def __setitem__(self, key, value):
        self.dirichlet[key] = value

    @classmethod
    def from_workspace(cls, workspace):
        """ Generate ElasticityBC from a Workspace.

        :param workspace: domain
        :type workspace: Workspace
        """
        if isinstance(workspace, Workspace):
            return cls(workspace.len_x(), workspace.len_y(), workspace.len_z())
        else:
            raise Exception("Cannot create ElasticityBC, the input to the from_workspace method has to be a Workspace.")

    def get_shape(self):
        return self.dirichlet.shape

    def show(self):
        Workspace.show_orientation(self.dirichlet)



import os.path
import ntpath


def path_to_example_file(example_filename):
    """ Path to example data that is installed with pumapy (used for testing and tutorial)

    :param example_filename: name of the example file (check python/pumapy/data for a list of example files)
    :type example_filename: str
    :return: path to the example file, which can be used to import it
    :rtype: str

    :Example:
    >>> import pumapy as puma
    >>> ws_example = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif")) # import example file
    >>> puma.plot_slices(ws_example) # visualize example file
    """

    path = ntpath.split(os.path.dirname(os.path.realpath(__file__)))[0]
    file_path = os.path.join(path, 'data', example_filename)

    if not os.path.isfile(file_path):
        raise Exception("Example file not found.")

    return file_path


def list_example_files():
    """ List all example files available inside the folder python/pumapy/data

    :return: List of example file names
    :rtype: list
    """

    path = ntpath.split(os.path.dirname(os.path.realpath(__file__)))[0]
    file_path = os.path.join(path, 'data')

    file_list = os.listdir(file_path)
    return_list = []
    for file in file_list:
        if file[-4:] == ".tif" or file[-7:] == ".pumapy" or file[-4:] == ".vti":
            return_list.append(file)

    return return_list
